#!/usr/sbin/dtrace -s

/***************************************************************************
 * geli-key-monitor.d
 *
 * Traces geli to print the beginning of various keys.
 * Warns about already known keys.
 *
 * Copyright (c) 2012 Fabian Keil <fk@fabiankeil.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 ***************************************************************************/

#pragma D option quiet
#pragma D option dynvarsize=10m

dtrace:::BEGIN
{
        algo[ 2] = "3DES-CBC";
        algo[ 3] = "Blowfish-CBC";
        algo[11] = "AES-CBC";
        algo[21] = "Camellia-CBC";
        algo[22] = "AES-XTS";

        /*
         * These are the first bytes of known weak keys generated with
         * an unitialized sc->sc_ekey on a little-endian system. 
         */
        known_keys[0x06d800ef] = 1;
        known_keys[0x252c6a6d] = 1;
        known_keys[0xa72b9c7c] = 1;
        known_keys[0xfde44683] = 1;

        G_ELI_FLAG_ENC_IVKEY  = 0x00400000;

        g_eli_key_stats_available = 0;
        g_eli_hold_stats_available = 0;

        /*
         * Geli v5 and higher use a different key for each GB,
         * so large disks have several thousand keys.
         * Monitoring only the first few should do, keeps the
         * output and reduces the likelihood of prevents dynvar
         * drops.
         */
        max_disk_keys_to_monitor = 4;

        printf("%Y: Monitoring geli keys (up to %d encryption keys per disk). %s\n",
               walltimestamp, max_disk_keys_to_monitor, "Press CTRL-C to exit.");
}

fbt::g_eli_mkey_propagate:entry
{
        self->sc = (struct g_eli_softc *)arg0;
        self->mkey = (struct g_eli_key *)arg1;

        self->geom_name = stringof(self->sc->sc_geom->name);
        self->monitored_keys = 0;
        self->key_limit_reached = 0;

        this->algo = self->sc->sc_ealgo;
        this->algo_name = (algo[this->algo] != NULL) ? algo[this->algo] : "Unregistered";
        this->geli_version = self->sc->sc_version;

        printf("%Y: %s: %s:%-6s: version: %d, algo: %s, flags: 0x%x (FLAG_ENC_IVKEY: %d) ",
               walltimestamp,
               self->geom_name,
               probefunc, probename,
               this->geli_version, this->algo_name,
               self->sc->sc_flags,
               (self->sc->sc_flags & G_ELI_FLAG_ENC_IVKEY) != 0
               );
        printf("mkey: %08.8x, sc_mkey: %08.8x, sc_ekey: %08.8x\n",
               *(uint32_t *)(self->mkey),
               *(uint32_t *)(self->sc->sc_mkey),
               *(uint32_t *)(self->sc->sc_ekey));
}

fbt::g_eli_mkey_propagate:return
/self->sc != NULL/
{
        printf("%Y: %s: %s:%-6s: mkey: %08.8x, sc_mkey: %08.8x, sc_ekey: %08.8x, sc_ekeys_allocated: %d\n",
               walltimestamp,
               self->geom_name,
               probefunc, probename,
               *(uint32_t *)(self->mkey),
               *(uint32_t *)(self->sc->sc_mkey),
               *(uint32_t *)(self->sc->sc_ekey),
               self->sc->sc_ekeys_allocated);
}

fbt::g_eli_key_fill:entry
{
        self->sc    = (struct g_eli_softc *)arg0;
        self->key   = (struct g_eli_key *)arg1;
        self->geom_name = stringof(self->sc->sc_geom->name);
}

fbt::g_eli_key_fill:
/(self->sc != NULL) && !self->key_limit_reached/
{
        this->key = *(uint32_t *)(self->key->gek_key);
        printf("%Y: %s: %s:%-6s: key->gek_key: %8.8x (%d).%s\n",
               walltimestamp,
               self->geom_name,
               probefunc, probename,
               this->key,
               self->monitored_keys,
               known_keys[this->key] ? " Key looks familiar!" : "");
}

fbt::g_eli_key_fill:return
/(self->sc != NULL) && !self->key_limit_reached/
{
        this->key = *(uint32_t *)(self->key->gek_key);
        @g_eli_keys[self->monitored_keys, this->key, self->geom_name] = count();
        /*
         * Register the generated key as known.
         * Unless the provider is reattached we do not want to see it again.
         */
        known_keys[this->key] = 1;
        g_eli_key_stats_available = 1;
        self->monitored_keys++;
}

fbt::g_eli_key_fill:return
/(self->sc != NULL) && (self->sc->sc_ekeys_allocated == max_disk_keys_to_monitor - 1)/
{
        printf("%Y: %s: %s:%-6s: Encryption key limit per disk reached.\n",
               walltimestamp, self->geom_name, probefunc, probename);
        self->key_limit_reached = 1;
}

fbt::g_eli_key_hold:entry
{
        self->sc        = (struct g_eli_softc *)arg0;
        self->offset    = (off_t)arg1;
        self->blocksize = (size_t)arg2;
        self->geom_name = stringof(self->sc->sc_geom->name);

        this->first_mkey_bytes = *(uint32_t *)(self->sc->sc_mkey);
        this->first_ekey_bytes = *(uint32_t *)(self->sc->sc_ekey);
        /*
         *  We only check the first bytes here, so false-positives are
         *  theoretically posible although unlikely.
         */
        this->empty_ekey = (0 == this->first_ekey_bytes);

        this->geli_version = self->sc->sc_version;
        this->algo      = self->sc->sc_ealgo;
        this->algo_name = (algo[this->algo] != NULL) ? algo[this->algo] : "Unregistered";

        @g_eli_key_hold[self->geom_name,
                        this->geli_version,
                        this->algo_name,
                        this->algo,
                        this->first_mkey_bytes,
                        this->first_ekey_bytes] = count();

        g_eli_hold_stats_available = 1;
}

tick-60sec,
dtrace:::END
/g_eli_hold_stats_available/
{
        printf("\n---------------------------------------------------------\n");
        printf("%Y: g_eli_key_hold() calls so far:\n", walltimestamp);
        printf("%-20s %12s %18s %11s %11s %10s\n",
               "Provider", "Geli version", "Algorithm", "mkey start", "ekey start", "calls");
        printa("%-20s %12d %13s (%2d)    %08x    %08x %@10d\n", @g_eli_key_hold);
}

tick-60sec,
dtrace:::END
/g_eli_key_stats_available/
{
        printf("\n%Y: g_eli_key count (only works for geli version 5 or higher):\n", walltimestamp);
        printa("%@u #%04d %8.8x %-25s\n", @g_eli_keys);
}
